import logging
import os
from typing import Any, TextIO

import pandas as pd

from otoole.input import WriteStrategy
from otoole.read_strategies import CSV_TO_EXCEL
from otoole.utils import read_packaged_file

logger = logging.getLogger(__name__)


class WriteExcel(WriteStrategy):
    def _header(self):
        return pd.ExcelWriter(self.filepath, mode="w")

    def _form_parameter(
        self, df: pd.DataFrame, parameter_name: str, default: float
    ) -> pd.DataFrame:
        """Converts data into wide format

        Arguments
        ---------
        df: pd.DataFrame
        parameter_name: str
        default: float

        Returns
        -------
        pandas.DataFrame
        """

        if not df.empty:

            names = df.columns.to_list()
            if len(names) > 2:
                logger.debug(
                    "More than 2 columns for {}: {}".format(parameter_name, names)
                )
                rows = names[0:-2]
                columns = names[-2]
                values = names[-1]
                logger.debug("Rows: %s; columns: %s; values: %s", rows, columns, values)
                logger.debug("dtypes: {}".format(df.dtypes))
                pivot = pd.pivot_table(
                    df, index=rows, columns=columns, values=values, fill_value=default
                )
            elif len(names) == 2:
                logger.debug("Two columns for {}: {}".format(parameter_name, names))
                rows = names[0]
                values = names[1]
                logger.debug("Rows: %s; values: %s", rows, values)
                pivot = pd.pivot_table(
                    df, index=rows, values=values, fill_value=default
                )
            else:
                logger.debug("One column for {}: {}".format(parameter_name, names))
                pivot = df.copy()
                pivot = pivot.reset_index(drop=True)

        else:
            logger.debug("Dataframe {} is empty".format(parameter_name))
            pivot = df.copy()

        return pivot

    def _write_parameter(
        self,
        df: pd.DataFrame,
        parameter_name: str,
        handle: pd.ExcelWriter,
        default: float,
    ):
        try:
            name = CSV_TO_EXCEL[parameter_name]
        except KeyError:
            name = parameter_name
        df = self._form_parameter(df, parameter_name, default)
        df.to_excel(handle, sheet_name=name, merge_cells=False)

    def _write_set(self, df: pd.DataFrame, set_name, handle: pd.ExcelWriter):
        df.to_excel(handle, sheet_name=set_name, merge_cells=False, index=False)

    def _footer(self, handle=pd.ExcelWriter):
        handle.close()


class WriteDatafile(WriteStrategy):
    def _header(self):
        filepath = open(self.filepath, "w", newline="")
        msg = "# Model file written by *otoole*\n"
        filepath.write(msg)
        return filepath

    def _form_parameter(self, df: pd.DataFrame, default: float):

        df = df[df.VALUE != default]
        return df

    def _write_parameter(
        self, df: pd.DataFrame, parameter_name: str, handle: TextIO, default: float
    ):
        """Write parameter data to a GMPL datafile, omitting data with default value

        Arguments
        ---------
        filepath : StreamIO
        df : pandas.DataFrame
        parameter_name : str
        handle: TextIO
        default : int
        """
        df = self._form_parameter(df, default)
        handle.write("param default {} : {} :=\n".format(default, parameter_name))
        df.to_csv(path_or_buf=handle, sep=" ", header=False, index=True)
        handle.write(";\n")

    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO):
        """Write set data to a GMPL datafile

        Arguments
        ---------
        df : pandas.DataFrame
        set_name : str
        handle: TextIO
        """
        handle.write("set {} :=\n".format(set_name))
        df.to_csv(path_or_buf=handle, sep=" ", header=False, index=False)
        handle.write(";\n")

    def _footer(self, handle: TextIO):
        handle.write("end;\n")
        handle.close()


class WriteCsv(WriteStrategy):
    """Write parameters to comma-separated value files

    Arguments
    ---------
    filepath: str, default=None
        The path to write a folder of csv files
    default_values: dict, default=None
    user_config: dict, default=None
    """

    def _header(self) -> Any:
        os.makedirs(os.path.join(self.filepath), exist_ok=True)
        return None

    def _write_parameter(
        self, df: pd.DataFrame, parameter_name: str, handle: TextIO, default: float
    ) -> pd.DataFrame:
        """Write parameter data"""
        self._write_out_dataframe(self.filepath, parameter_name, df)

    def _write_out_dataframe(self, folder, parameter, df):
        """Writes out a dataframe as a csv into the data subfolder of a datapackage

        Arguments
        ---------
        folder : str
        parameter : str
        df : pandas.DataFrame

        """
        filepath = os.path.join(folder, parameter + ".csv")
        with open(filepath, "w", newline="") as csvfile:
            logger.info(
                "Writing %s rows into narrow file for %s", df.shape[0], parameter
            )
            df.to_csv(csvfile, index=False)

    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO) -> pd.DataFrame:
        """Write set data"""
        self._write_out_dataframe(self.filepath, set_name, df)

    def _footer(self, handle: TextIO):
        pass


class WriteDatapackage(WriteCsv):
    def _header(self) -> Any:
        os.makedirs(os.path.join(self.filepath, "data"), exist_ok=True)
        return None

    def _write_out_dataframe(self, folder, parameter, df):
        """Writes out a dataframe as a csv into the data subfolder of a datapackage

        Arguments
        ---------
        folder : str
        parameter : str
        df : pandas.DataFrame

        """
        filepath = os.path.join(folder, "data", parameter + ".csv")
        with open(filepath, "w", newline="") as csvfile:
            logger.info(
                "Writing %s rows into narrow file for %s", df.shape[0], parameter
            )
            df.to_csv(csvfile, index=False)

    def _footer(self, handle: TextIO):
        datapackage = read_packaged_file("datapackage.json", "otoole.preprocess")
        filepath = os.path.join(self.filepath, "datapackage.json")
        with open(filepath, "w", newline="") as destination:
            destination.writelines(datapackage)
        self._write_default_values()

    def _write_default_values(self):

        default_values_path = os.path.join(self.filepath, "data", "default_values.csv")
        with open(default_values_path, "w", newline="") as csv_file:
            csv_file.write("name,default_value\n")

            for name, contents in self.input_config.items():
                if contents["type"] == "param":
                    csv_file.write("{},{}\n".format(name, contents["default"]))



"""Visualise the reference energy system
"""
import logging
import os
import sys
from typing import Dict, List, Tuple

import networkx as nx  # mypy: ignore
from datapackage import Package

logger = logging.getLogger(__name__)


def load_datapackage(path_to_datapackage: str) -> Package:

    package = Package(path_to_datapackage)

    return package


def extract_nodes(
    package_rows: List[List], node_type="technology", color="red", shape="circle"
) -> List[Tuple[str, Dict]]:
    """Add nodes from a Tabular Data Table

    Arguments
    ---------
    package_rows : list of dict
    node_type : str, default='technology'
    color : str, default='red'
    shape : str, default='circle'

    Returns
    -------
    list of tuple
        A list of nodes with attributes
    """
    nodes = [
        (
            x[0],
            {
                "type": node_type,
                "fillcolor": color,
                "shape": shape,
                "style": "filled",
                "label": x[0],
            },
        )
        for x in package_rows
    ]

    return nodes


def add_fuel(package_rows: List[List]) -> List[Tuple[str, Dict]]:
    """Add fuel nodes

    Arguments
    ---------
    package_rows : list of dict

    Returns
    -------
    list of tuple
        A list of node names along with a dict of node attributes
    """
    nodes = [
        (
            x[0],
            {
                "type": "fuel",
                "style": "filled",
                "shape": "circle",
                "height": 0.1,
                "width": 0.1,
                "label": "",
            },
        )
        for x in package_rows
    ]
    return nodes


def extract_edges(
    package_rows: List[Dict],
    from_column: str,
    to_column: str,
    parameter_name: str,
    directed: bool = True,
) -> List[Tuple[str, str, Dict]]:
    """Add edges from a Tabular Data Table

    Arguments
    ---------
    package_rows : list of dict
    from_column : str
        The name of the column to use as source of the edge
    to_column: str
        The name of the column to use as a destination of the edge
    parameter_name: str
        The name of the parameter
    directed: bool, default=True
        Specifies whether the edge should have an arrow or not

    Returns
    -------
    list of tuple
        A list of edges with from/to nodes names and edge attributes
    """
    if directed:
        edges = [
            (
                x[from_column],
                x[to_column],
                {parameter_name: float(x["VALUE"]), "dir": "none"},
            )
            for x in package_rows
        ]
    else:
        edges = [
            (
                x[from_column],
                x[to_column],
                {parameter_name: float(x["VALUE"]), "xlabel": x[from_column]},
            )
            for x in package_rows
        ]

    return edges


def create_graph(datapackage: Package):
    """Creates a graph of technologies and fuels

    Arguments
    ---------
    datapackage : datapackage.Package

    Returns
    -------
    graph
        networkx.DiGraph
    """

    technologies = datapackage.get_resource("TECHNOLOGY").read()
    storage = datapackage.get_resource("STORAGE").read()
    fuel = datapackage.get_resource("FUEL").read()
    emission = datapackage.get_resource("EMISSION").read()

    nodes = extract_nodes(technologies, shape="rectangle", color="yellow")
    nodes += extract_nodes(
        storage, node_type="storage", shape="triangle", color="lightblue"
    )
    nodes += add_fuel(fuel)
    nodes += extract_nodes(emission, node_type="emission", color="grey")
    nodes += [
        (
            "AnnualDemand",
            {
                "type": "demand",
                "fillcolor": "green",
                "label": "AccumulatedAnnualDemand",
                "style": "filled",
            },
        )
    ]

    input_activity = datapackage.get_resource("InputActivityRatio").read(keyed=True)
    output_activity = datapackage.get_resource("OutputActivityRatio").read(keyed=True)
    emission_activity = datapackage.get_resource("EmissionActivityRatio").read(
        keyed=True
    )
    tech2storage = datapackage.get_resource("TechnologyToStorage").read(keyed=True)
    techfromstorage = datapackage.get_resource("TechnologyFromStorage").read(keyed=True)
    acc_demand = datapackage.get_resource("AccumulatedAnnualDemand").read(keyed=True)
    spec_demand = datapackage.get_resource("SpecifiedAnnualDemand").read(keyed=True)

    edges = extract_edges(
        input_activity, "FUEL", "TECHNOLOGY", "input_ratio", directed=False
    )

    edges += extract_edges(output_activity, "TECHNOLOGY", "FUEL", "output_ratio")

    edges += extract_edges(
        emission_activity, "TECHNOLOGY", "EMISSION", "emission_ratio"
    )

    edges += extract_edges(tech2storage, "TECHNOLOGY", "STORAGE", "input_ratio")

    edges += extract_edges(techfromstorage, "STORAGE", "TECHNOLOGY", "ouput_ratio")

    edges += [
        (x["FUEL"], "AnnualDemand", {"Demand": float(x["VALUE"]), "xlabel": x["FUEL"]})
        for x in acc_demand
    ]
    edges += [
        (x["FUEL"], "AnnualDemand", {"Demand": float(x["VALUE"]), "xlabel": x["FUEL"]})
        for x in spec_demand
    ]

    graph = build_graph(nodes, edges)

    return graph


def create_res(path_to_datapackage: str, path_to_resfile: str):
    """Create a reference energy system diagram from a Tabular Data Package

    Arguments
    ---------
    path_to_datapackage : str
        The path to the ``datapackage.json``
    path_to_resfile : str
        The path to the image file to be created
    """
    package = load_datapackage(path_to_datapackage)

    graph = create_graph(package)
    draw_graph(graph, path_to_resfile)


def draw_graph(graph, path_to_resfile):
    """Layout the graph and write it to disk

    Uses pygraphviz to set some graph attributes, layout the graph
    and write it to disk

    Arguments
    ---------
    path_to_resfile : str
        The file path of the PNG image file that will be created
    """
    for node, attributes in graph.nodes.data():
        logger.debug("%s: %s", node, attributes)
    for source, sink, attributes in graph.edges.data():
        logger.debug("%s-%s: %s", source, sink, attributes)

    filename, ext = os.path.splitext(path_to_resfile)
    nx.write_graphml(graph, filename + ".graphml")
    dot_graph = nx.nx_pydot.to_pydot(graph)

    dot_graph.set("rankdir", "LR")
    dot_graph.set("splines", "ortho")
    dot_graph.set("concentrate", "true")

    image_format = ext.strip(".")

    image = dot_graph.create(prog="dot", format=image_format)
    with open(path_to_resfile, "wb") as image_file:
        image_file.write(image)


def build_graph(
    nodes: List[Tuple[str, Dict]], edges: List[Tuple[str, str, Dict]]
) -> nx.DiGraph:
    """Builds the graph using networkx

    Arguments
    ---------
    nodes : list
        A list of node tuples
    edges : list
        A list of edge tuples

    Returns
    -------
    networkx.DiGraph
        A directed graph representing the reference energy system
    """
    graph = nx.DiGraph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    return graph


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG)
    path_to_datapackage = sys.argv[1]
    path_to_resfile = sys.argv[2]
    create_res(path_to_datapackage, path_to_resfile)



import os
from subprocess import run
from tempfile import NamedTemporaryFile, mkdtemp

from pytest import mark

from otoole import __version__


class TestConvert:
    def test_version(self):
        result = run(["otoole", "--version"], capture_output=True)
        assert result.stdout.strip().decode() == str(__version__)

    temp = mkdtemp()
    temp_excel = NamedTemporaryFile(suffix="xlsx")
    temp_datafile = NamedTemporaryFile(suffix="dat")
    simplicity = os.path.join("tests", "fixtures", "simplicity.txt")

    test_data = [
        (["otoole", "convert", "--help"], "usage: otoole convert [-h]"),
        (["otoole", "convert", "datafile", "datapackage", simplicity, temp], ""),
        (["otoole", "convert", "datafile", "excel", simplicity, temp_excel.name], ""),
        (
            [
                "otoole",
                "convert",
                "datafile",
                "datafile",
                simplicity,
                temp_datafile.name,
            ],
            "",
        ),
    ]

    @mark.parametrize(
        "commands,expected", test_data, ids=["help", "datapackage", "excel", "datafile"]
    )
    def test_convert_commands(self, commands, expected):
        actual = run(commands, capture_output=True)
        assert expected in str(actual.stdout)

    test_errors = [
        (
            ["otoole", "convert", "datafile", "cplex", "x", "y"],
            "error: argument to_format: invalid choice: 'cplex'",
        )
    ]

    @mark.parametrize("commands,expected", test_errors, ids=["invalid"])
    def test_convert_error(self, commands, expected):
        actual = run(commands, capture_output=True)
        assert expected in str(actual.stderr)



import io

import pandas as pd

from otoole.write_strategies import WriteDatafile, WriteExcel


class TestWriteExcel:
    def test_form_empty_parameter_with_defaults(self):

        convert = WriteExcel()  # typing: WriteExcel

        data = []

        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "VALUE"])
        actual = convert._form_parameter(df, "test_parameter", 0)
        expected = pd.DataFrame(data=data, columns=["REGION", "FUEL", "VALUE"])
        pd.testing.assert_frame_equal(actual, expected)

    def test_form_empty_two_index_param_with_defaults(self):

        convert = WriteExcel()  # typing: WriteExcel

        df = pd.DataFrame(data=[], columns=["REGION", "VALUE"])
        actual = convert._form_parameter(df, "test_parameter", 0)
        expected = pd.DataFrame(data=[], columns=["REGION", "VALUE"])
        pd.testing.assert_frame_equal(actual, expected)

    def test_form_two_index_param(self):

        convert = WriteExcel()  # typing: WriteExcel

        df = pd.DataFrame(
            data=[["SIMPLICITY", 0.10], ["UTOPIA", 0.20]], columns=["REGION", "VALUE"]
        )
        actual = convert._form_parameter(df, "test_parameter", 0)
        index = pd.Index(data=["SIMPLICITY", "UTOPIA"], name="REGION")
        expected = pd.DataFrame(data=[[0.10], [0.20]], columns=["VALUE"], index=index)
        pd.testing.assert_frame_equal(actual, expected)

    def test_form_one_columns(self):

        convert = WriteExcel()  # typing: WriteExcel

        data = ["A", "B", "C"]

        df = pd.DataFrame(data=data, columns=["FUEL"])
        actual = convert._form_parameter(df, "test_set", 0)
        expected = pd.DataFrame(data=["A", "B", "C"], columns=["FUEL"])
        pd.testing.assert_frame_equal(actual, expected)

    def test_form_three_columns(self):

        convert = WriteExcel()  # typing: WriteExcel

        data = [["SIMPLICITY", "COAL", 2015, 41], ["SIMPLICITY", "COAL", 2016, 42]]

        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "YEAR", "VALUE"])
        actual = convert._form_parameter(df, "test_parameter", 0)

        expected_data = [[41, 42]]
        expected = pd.DataFrame(
            data=expected_data,
            columns=pd.Int64Index([2015, 2016], dtype="int64", name="YEAR"),
            index=pd.MultiIndex.from_tuples(
                [("SIMPLICITY", "COAL")], names=["REGION", "FUEL"]
            ),
        )

        pd.testing.assert_frame_equal(actual, expected)


class TestWriteDatafile:
    def test_write_empty_parameter_with_defaults(self):

        convert = WriteDatafile()  # typing: WriteDatafile

        data = []

        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "VALUE"])

        stream = io.StringIO()
        convert._write_parameter(df, "test_parameter", stream, 0)

        stream.seek(0)
        expected = ["param default 0 : test_parameter :=\n", ";\n"]
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line

    def test_write_parameter_as_tabbing_format(self):

        data = [["SIMPLICITY", "BIOMASS", 0.95969], ["SIMPLICITY", "ETH1", 4.69969]]

        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "VALUE"]).set_index(
            ["REGION", "FUEL"]
        )

        stream = io.StringIO()
        convert = WriteDatafile()
        convert._write_parameter(df, "test_parameter", stream, 0)

        stream.seek(0)
        expected = [
            "param default 0 : test_parameter :=\n",
            "SIMPLICITY BIOMASS 0.95969\n",
            "SIMPLICITY ETH1 4.69969\n",
            ";\n",
        ]
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line

    def test_write_parameter_skip_defaults(self):

        data = [
            ["SIMPLICITY", "BIOMASS", 0.95969],
            ["SIMPLICITY", "ETH1", 4.69969],
            ["SIMPLICITY", "ETH2", -1],
            ["SIMPLICITY", "ETH3", -1],
        ]

        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "VALUE"]).set_index(
            ["REGION", "FUEL"]
        )

        stream = io.StringIO()
        convert = WriteDatafile()
        convert._write_parameter(df, "test_parameter", stream, -1)

        stream.seek(0)
        expected = [
            "param default -1 : test_parameter :=\n",
            "SIMPLICITY BIOMASS 0.95969\n",
            "SIMPLICITY ETH1 4.69969\n",
            ";\n",
        ]
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line

    def test_write_set(self):

        data = [["BIOMASS"], ["ETH1"]]

        df = pd.DataFrame(data=data, columns=["VALUE"])

        stream = io.StringIO()
        convert = WriteDatafile()
        convert._write_set(df, "TECHNOLOGY", stream)

        stream.seek(0)
        expected = ["set TECHNOLOGY :=\n", "BIOMASS\n", "ETH1\n", ";\n"]
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line



"""

read_strategy = None
write_strategy = None

if args.from_format == "datafile":
    read_strategy = ReadDatafile()
elif args.from_format == "datapackage":
    read_strategy = ReadDatapackage()
elif args.from_format == "csv":
    read_strategy = ReadCsv()
elif args.from_format == "excel":
    read_strategy = ReadExcel()

if args.to_format == "datapackage":
    write_strategy = WriteDatapackage()
elif args.to_format == "excel":
    write_strategy = WriteExcel()
elif args.to_format == "datafile":
    write_strategy = WriteDatafile()
elif args.to_format == "csv":
    write_strategy = WriteCsv()

if read_strategy and write_strategy:
    context = Context(read_strategy, write_strategy)
    context.convert(args.from_path, args.to_path)
else:
    raise NotImplementedError(msg)

"""
import os
from tempfile import NamedTemporaryFile

from otoole import Context, ReadExcel, WriteDatafile


class TestConvert:
    def test_convert_excel_to_datafile(self):

        read_strategy = ReadExcel()
        write_strategy = WriteDatafile()
        context = Context(read_strategy, write_strategy)

        tmpfile = NamedTemporaryFile()
        from_path = os.path.join("tests", "fixtures", "combined_inputs.xlsx")

        context.convert(from_path, tmpfile.name)

        tmpfile.seek(0)
        actual = tmpfile.readlines()
        tmpfile.close()

        assert actual[-1] == b"end;\n"
        assert actual[0] == b"# Model file written by *otoole*\n"
        assert actual[2] == b"09_ROK d_bld_2_coal_products 2017 20.892132\n"
        assert actual[8996] == b"param default 1 : DepreciationMethod :=\n"
